#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <stdarg.h>
#include <limits.h>
#include <errno.h>

#include "srv.h"

#define SUPPORTSRELOAD_DEFAULT 0
#define NICE_DEFAULT 0
#define SEQUENCE_DEFAULT 20

char *language = NULL;
char *language2 = NULL;

void message(const char *format, ...)
{
	va_list ap;

	va_start(ap, format);
	vfprintf(stderr, format, ap);
	va_end(ap);
	fprintf(stderr, "\n");
}

void message_e(const char *format, ...)
{
	va_list ap;
	int saved_errno = errno;

	va_start(ap, format);
	vfprintf(stderr, format, ap);
	va_end(ap);
	fprintf(stderr, ": %s\n", strerror(saved_errno));
}

static int string_to_int (const char *s, int err)
{
	int val;
	char dummy;
	if (sscanf(s, "%d%c", &val, &dummy) != 1) return err;
	return val;
}

static int script_add (const char *lang, const char *value, struct service *service)
{
	if (value[0] != '\0' && value[0] != '/') {
		message("script has to be given with absolute pathname");
		return -1;
	}
	if (service->script != NULL) free(service->script);
	service->script = strdup(value);
	return 0;
}

static int script_remove (const char *lang, struct service *service)
{
	if (service->script != NULL) free(service->script);
	service->script = NULL;
	return 0;
}

static int script_dump (const struct service *service, FILE *f)
{
	if (service->script != NULL) {
		fprintf(f, "Script=%s\n", service->script);
	}
	return 0;
}

static int supportsreload_add (const char *lang, const char *value, struct service *service)
{
	service->supportsreload_type = givenvalue;
	if (!strcasecmp(value, "yes") || !strcasecmp(value, "true") || !strcmp(value, "1")) {
		service->supportsreload = 1;
	} else if (!strcasecmp(value, "no") || !strcasecmp(value, "false") || !strcmp(value, "0")) {
		service->supportsreload = 0;
	} else {
		message("invalid value for SupportsReload");
		return -1;
	}
	return 0;
}

static int supportsreload_remove (const char *lang, struct service *service)
{
	service->supportsreload_type = unaltered;
	return 0;
}

static int supportsreload_dump (const struct service *service, FILE *f)
{
	if (service->supportsreload_type == cleared) {
		fprintf(f, "SupportsReload=\n");
	} else if (service->supportsreload_type == givenvalue) {
		fprintf(f, "SupportsReload=%s\n", service->supportsreload ? "yes" : "no");
	}
	return 0;
}

static int runlevels_add (const char *lang, const char *value, struct service *service)
{
	int i;
	if (strspn(value, "0123456") != strlen(value)) {
		message("invalid value for Runlevels");
		return -1;
	}
	for (i = 0; i <= 6; i++) {
		service->runlevel[i] = (strchr(value, '0' + i) == NULL) ? 0 : 1;
	}
	service->runlevel_type = givenvalue;
	return 0;
}

static int runlevels_remove (const char *lang, struct service *service)
{
	service->runlevel_type = unaltered;
	return 0;
}

static int runlevels_dump (const struct service *service, FILE *f)
{
	int i;
	if (service->runlevel_type == givenvalue) {
		fprintf(f, "Runlevels=");
		for (i = 0; i <= 6; i++) {
			if (service->runlevel[i]) fputc('0' + i, f);
		}
		fprintf(f, "\n");
	}
	return 0;
}

static int sequence_add (const char *lang, const char *value, struct service *service)
{
	int i;
	if (value[0] == '\0') {
		service->sequence_type = cleared;
	} else {
		i = string_to_int(value, 100);
		if (i < 0 || i > 99) {
			message("invalid sequence value");
			return -1;
		}
		service->sequence_type = givenvalue;
		service->sequence = i;
	}
	return 0;
}

static int sequence_remove (const char *lang, struct service *service)
{
	service->sequence_type = unaltered;
	return 0;
}

static int sequence_dump (const struct service *service, FILE *f)
{
	if (service->sequence_type == cleared) {
		fprintf(f, "Sequence=\n");
	} else if (service->sequence_type == givenvalue) {
		fprintf(f, "Sequence=%02d\n", service->sequence);
	}
	return 0;
}

static int nice_add (const char *lang, const char *value, struct service *service)
{
	int i;
	if (value[0] == '\0') {
		service->nice_type = cleared;
	} else {
		i = string_to_int(value, 100);
		if (i < -20 || i > 19) {
			message("invalid nice value");
			return -1;
		}
		service->nice_type = givenvalue;
		service->nice = i;
	}
	return 0;
}

static int nice_remove (const char *lang, struct service *service)
{
	service->nice_type = unaltered;
	return 0;
}

static int nice_dump (const struct service *service, FILE *f)
{
	if (service->nice_type == cleared) {
		fprintf(f, "Nice=\n");
	} else if (service->nice_type == givenvalue) {
		fprintf(f, "Nice=%d\n", service->nice);
	}
	return 0;
}

static int localestring_add (struct localestring *ls, const char *lang, const char *value)
{
	int i = 0;
	if (lang == NULL) lang = "";
	while (i < ls->nr) {
		if (!strcmp(ls->langvalues[i].lang, lang)) {
			free(ls->langvalues[i].value);
			ls->langvalues[i].value = strdup(value);
			return 0;
		}
		i++;
	}
	ls->nr++;
	ls->langvalues = realloc(ls->langvalues, ls->nr * sizeof(struct langvalue));
	ls->langvalues[ls->nr-1].lang = strdup(lang);
	ls->langvalues[ls->nr-1].value = strdup(value);
	return 0;
}

static int localestring_remove (struct localestring *ls, const char *lang)
{
	int i = 0, j;
	if (lang == NULL) lang = "";
	while (i < ls->nr) {
		if (!strcmp(ls->langvalues[i].lang, lang)) {
			free(ls->langvalues[i].lang);
			free(ls->langvalues[i].value);
			j = i;
			while (j < ls->nr - 1) {
				memcpy(&ls->langvalues[j], &ls->langvalues[j+1], sizeof(struct langvalue));
				j++;
			}
			ls->nr--;
			ls->langvalues = realloc(ls->langvalues, ls->nr * sizeof(struct langvalue));
			return 0;
		}
		i++;
	}
	return 0;
}

static int localestring_dump (const struct localestring *ls, const char *name, FILE *f)
{
	int i = 0;
	int l1 = -1, l2 = -1, l3 = -1;
	while (i < ls->nr) {
		if (language != NULL && !strcmp(ls->langvalues[i].lang, language)) l1 = i;
		if (language2 != NULL && !strcmp(ls->langvalues[i].lang, language2)) l2 = i;
		if (ls->langvalues[i].lang[0] == '\0') l3 = i;
		fprintf(f, "%s", name);
		if (ls->langvalues[i].lang[0] != '\0') fprintf(f, "[%s]", ls->langvalues[i].lang);
		fprintf(f, "=%s\n", ls->langvalues[i].value);
		i++;
	}
	if (language != NULL) {
		if (l1 >= 0) i = l1;
		else if (l2 >= 0) i = l2;
		else i = l3;
		if (i >= 0) {
			fprintf(f, "%s[]=%s\n", name, ls->langvalues[i].value);
		}
	}
	return 0;
}

static int summary_add (const char *lang, const char *value, struct service *service)
{
	return localestring_add(service->summary, lang, value);
}

static int summary_remove (const char *lang, struct service *service)
{
	return localestring_remove(service->summary, lang);
}

static int summary_dump (const struct service *service, FILE *f)
{
	return localestring_dump(service->summary, "Summary", f);
}

static int description_add (const char *lang, const char *value, struct service *service)
{
	return localestring_add(service->description, lang, value);
}

static int description_remove (const char *lang, struct service *service)
{
	return localestring_remove(service->description, lang);
}

static int description_dump (const struct service *service, FILE *f)
{
	return localestring_dump(service->description, "Description", f);
}

#define foo(x) x ## _add, x ## _remove, x ## _dump
static struct valid_keyword valid_keywords[] = { 
	{ "Summary",        1, foo(summary)        },
	{ "Description",    1, foo(description)    },
	{ "Script",         0, foo(script)         },
	{ "SupportsReload", 0, foo(supportsreload) },
	{ "Runlevels",      0, foo(runlevels)      },
	{ "Sequence",       0, foo(sequence)       },
	{ "Nice",           0, foo(nice)           },
	{ NULL,	            0, NULL, NULL, NULL    }
};
#undef foo

static struct localestring *new_localestring (void)
{
	struct localestring *ls = malloc(sizeof(struct localestring));
	ls->langvalues = NULL;
	ls->nr = 0;
	return ls;
}

static int service_init (struct service *service)
{
	service->script = NULL;
	service->supportsreload_type = unaltered;
	service->runlevel_type = unaltered;
	service->sequence_type = unaltered;
	service->nice_type = unaltered;
	service->summary = new_localestring();
	service->description = new_localestring();
	return 0;
}

static int localestring_compare (const void *v1, const void *v2)
{
	const struct langvalue *lv1 = v1;
	const struct langvalue *lv2 = v2;
	return strcmp(lv1->lang, lv2->lang);
}

static void localestring_sort (struct localestring *ls)
{
	qsort(ls->langvalues, ls->nr, sizeof(struct langvalue), localestring_compare);
}

static int service_fix (struct service *service)
{
	int i;
	if (service->supportsreload_type != givenvalue) service->supportsreload = SUPPORTSRELOAD_DEFAULT;
	if (service->sequence_type != givenvalue) service->sequence = SEQUENCE_DEFAULT;
	if (service->nice_type != givenvalue) service->nice = NICE_DEFAULT;
	if (service->runlevel_type != givenvalue) {
		for (i = 0; i <= 6; i++) {
			service->runlevel[i] = 0;
		}
	}
	localestring_sort(service->summary);
	localestring_sort(service->description);
	return 0;
}

int service_add3 (char *keyword, char *lang, char *value, struct service *service)
{
	struct valid_keyword *vk = valid_keywords;
	while (vk->name != NULL) {
		if (!strcmp(keyword, vk->name)) {
			if (lang != NULL && !vk->is_localestring) {
				message("keyword '%s' isn't locale dependent", keyword);
				return -1;
			}
			return (*vk->add_fn)(lang, value, service);
		}
		vk++;
	}
	message("invalid keyword '%s'", keyword);
	return -1;
}

int service_add2 (char *keywordlang, char *value, struct service *service)
{
	char *end = keywordlang + strlen(keywordlang);
	char *lang = strchr(keywordlang, '[');
	if (lang != NULL && *(end-1) == ']') {
		*lang++ = '\0';
		*(--end) = '\0';
		if (lang == end) {
			message("empty language");
			return -1;
		}
	}
	if (*keywordlang == '\0') {
		message("missing keyword");
		return -1;
	}
	return service_add3(keywordlang, lang, value, service);
}

int service_add1 (char *s, struct service *service)
{
	char *value;
	char *end = s + strlen(s);
	while (end > s && (*(end-1) == ' ' || *(end-1) == '\t')) *(--end) = '\0';
	while (*s == ' ' || *s == '\t') s++;
	if (*s == '\0' || *s == '#') return 0;
	value = strchr(s, '=');
	if (value == NULL) {
		message("missing '=' sign");
		return -1;
	}
	end = value;
	*end = '\0';
	value++;
	while (*value == ' ' || *value == '\t') value++;
	while (end > s && (*(end-1) == ' ' || *(end-1) == '\t')) *(--end) = '\0';
	return service_add2(s, value, service);
}

int service_remove2 (char *keyword, char *lang, struct service *service)
{
	struct valid_keyword *vk = valid_keywords;
	while (vk->name != NULL) {
		if (!strcmp(keyword, vk->name)) {
			if (lang != NULL && !vk->is_localestring) {
				message("keyword '%s' isn't locale dependent", keyword);
				return -1;
			}
			return (*vk->remove_fn)(lang, service);
		}
		vk++;
	}
	message("invalid keyword '%s'", keyword);
	return -1;
}

int service_remove1 (char *keywordlang, struct service *service)
{
	char *end = keywordlang + strlen(keywordlang);
	char *lang = strchr(keywordlang, '[');
	if (lang != NULL && *(end-1) == ']') {
		*lang++ = '\0';
		*(--end) = '\0';
		if (lang == end) {
			message("empty language");
			return -1;
		}
	}
	if (*keywordlang == '\0') {
		message("missing keyword");
		return -1;
	}
	return service_remove2(keywordlang, lang, service);
}

static int file_to_service (const char *filename, struct service *service)
{
	int fd;
	struct stat st;
	int size;
	char *content0, *content, *s;

	fd = open(filename, O_RDONLY);
	if (fd < 0) {
		return 0;
	}
	if (fstat(fd, &st) < 0) {
		message_e("fstat");
		close(fd);
		return -1;
	}
	size = st.st_size;
	content0 = malloc(size + 1);
	if (content0 == NULL) {
		message_e("malloc failed");
		close(fd);
		return -1;
	}

	if (read(fd, content0, size) != size) {
		message("read failed");
		close(fd);
		return -1;
	}
	close(fd);
	content0[size] = '\0';

	content = content0;
	while (content < content0 + size) {
		if (*content == '\n') *content = '\0';
		content++;
	}

	content = content0;
	while (content < content0 + size) {
		s = content + strlen(content) + 1;
		if (service_add1(content, service) < 0) return -1;
		content = s;
	}

	free(content0);
	return 0;
}

static int service_exists (const char *servicename)
{
	char buf[PATH_MAX];
	struct stat st;
	if (strchr(servicename, '/')) {
		return 0;
	}
	sprintf(buf, DEFAULTDIR "/%s.service", servicename);
	if (stat(buf, &st) == 0) return 1;
	sprintf(buf, CUSTOMDIR "/%s.service", servicename);
	if (stat(buf, &st) == 0) return 1;
	return 0;
}

int service_read_default (const char *servicename, struct service *service)
{
	char buf[PATH_MAX];
	if (!service_exists(servicename)) {
		message("invalid service name");
		return -1;
	}
	service_init(service);
	sprintf(buf, DEFAULTDIR "/%s.service", servicename);
	if (file_to_service(buf, service) < 0) return -1;
	return service_fix(service);
}

int service_read_custom (const char *servicename, struct service *service)
{
	char buf[PATH_MAX];
	if (!service_exists(servicename)) {
		message("invalid service name");
		return -1;
	}
	service_init(service);
	sprintf(buf, CUSTOMDIR "/%s.service", servicename);
	if (file_to_service(buf, service) < 0) return -1;
	return service_fix(service);
}

int service_read (const char *servicename, struct service *service)
{
	char buf[PATH_MAX];
	if (!service_exists(servicename)) {
		message("invalid service name");
		return -1;
	}
	service_init(service);
	sprintf(buf, DEFAULTDIR "/%s.service", servicename);
	if (file_to_service(buf, service) < 0) return -1;
	sprintf(buf, CUSTOMDIR "/%s.service", servicename);
	if (file_to_service(buf, service) < 0) return -1;
	return service_fix(service);
}

int service_dump (const struct service *service, FILE *f)
{
	struct valid_keyword *vk = &valid_keywords[0];
	while (vk->name != NULL) {
		(*vk->dump_fn)(service, f);
		vk++;
	}
	return 0;
}

static int string_compare (const void *v1, const void *v2)
{
	const char * const *s1 = v1;
	const char * const *s2 = v2;
	return strcmp(*s1, *s2);
}

int services_read_list (struct services *services)
{
	int i;
	DIR *d;
	struct dirent *de;
	services->name = NULL;
	services->nr = 0;
	d = opendir(DEFAULTDIR);
	if (d == NULL) {
		message_e("opendir() failed");
		return -1;
	}
	while ((de = readdir(d)) != NULL) {
		if (de->d_name[0] == '.') continue;
		if (strlen(de->d_name) < 9) continue;
		if (strcmp(de->d_name + strlen(de->d_name) - 8, ".service")) continue;
		de->d_name[strlen(de->d_name) - 8] = '\0';
		for (i = 0; i < services->nr; i++) {
			if (!strcmp(de->d_name, services->name[i])) break;
		}
		if (i < services->nr) continue;
		services->nr++;
		services->name = realloc(services->name, services->nr * sizeof(char *));
		services->name[services->nr-1] = strdup(de->d_name);
	}
	closedir(d);
	d = opendir(CUSTOMDIR);
	if (d == NULL) {
		message_e("opendir() failed");
		return -1;
	}
	while ((de = readdir(d)) != NULL) {
		if (de->d_name[0] == '.') continue;
		if (strlen(de->d_name) < 9) continue;
		if (strcmp(de->d_name + strlen(de->d_name) - 8, ".service")) continue;
		de->d_name[strlen(de->d_name) - 8] = '\0';
		for (i = 0; i < services->nr; i++) {
			if (!strcmp(de->d_name, services->name[i])) break;
		}
		if (i < services->nr) continue;
		services->nr++;
		services->name = realloc(services->name, services->nr * sizeof(char *));
		services->name[services->nr-1] = strdup(de->d_name);
	}
	closedir(d);
	qsort(services->name, services->nr, sizeof(char *), string_compare);
	return 0;
}

int services_dump_list (const struct services *services, FILE *f)
{
	int i;
	for (i = 0; i < services->nr; i++) {
		fprintf(f, "%s\n", services->name[i]);
	}
	return 0;
}

int use_current_language (void)
{
	char *s;
	if (language == NULL) language = getenv("LC_ALL");
	if (language == NULL) language = getenv("LC_MESSAGES");
	if (language == NULL) language = getenv("LANG");
	if (language == NULL) language = "";
	s = strchr(language, '@');
	if (s != NULL) *s = '\0';
	s = strchr(language, '.');
	if (s != NULL) *s = '\0';
	s = strchr(language, '_');
	if (s != NULL) {
		*s = '\0';
		language2 = strdup(language);
		*s = '_';
	}
	return 0;
}
